use crate::{Color, Shape, Shade, Value, Property};
use crate::error::Error;

/// A card in the game of set.
/// Internally, a set card is simply just a [`u8`]; an unsigned u8 integer.
#[derive(Clone, Copy)]
pub struct SetCard(u8);

impl_bit_operations!(Color);
impl_eq!(Color, Property::Color);
impl_bit_operations!(Shade);
impl_eq!(Shade, Property::Shade);
impl_bit_operations!(Shape);
impl_eq!(Shape, Property::Shape);
impl_bit_operations!(Value);
impl_eq!(Value, Property::Value);

impl TryFrom<u8> for SetCard {
  type Error = Error;

  fn try_from(value: u8) -> Result<Self, Self::Error> {
    println!("{:?}", value);
    if value & Property::Color == 0 {
      return Err(Error::NoCardPropertyDefined(Property::Color));
    }
    if value & Property::Shape == 0 {
      return Err(Error::NoCardPropertyDefined(Property::Shape));
    }
    if value & Property::Shade == 0 {
      return Err(Error::NoCardPropertyDefined(Property::Shade));
    }
    if value & Property::Value == 0 {
      return Err(Error::NoCardPropertyDefined(Property::Value));
    }
    Ok(Self(value))
  }
}

impl std::ops::BitAnd<Property> for SetCard {
  type Output = u8;

  fn bitand(self, rhs: Property) -> u8 {
    self.0 & (rhs as u8)
  }
}
impl std::ops::BitAnd<Property> for &SetCard {
  type Output = u8;

  fn bitand(self, rhs: Property) -> u8 {
    self.0 & (rhs as u8)
  }
}
impl std::ops::BitOr<Property> for SetCard {
  type Output = u8;

  fn bitor(self, rhs: Property) -> u8 {
    self.0 | (rhs as u8)
  }
}
impl std::ops::BitOr<Property> for &SetCard {
  type Output = u8;

  fn bitor(self, rhs: Property) -> u8 {
    self.0 | (rhs as u8)
  }
}

#[cfg(test)]
mod tests {
  use crate::{SetCard, Color, Shape, Shade, Value};

  #[test]
  fn test_good_card() {
    let _card = SetCard::try_from(Color::Red | Shape::Diamond | Shade::Empty | Value::Two).unwrap();
  }
  #[test]
  #[should_panic]
  fn test_missing_shape() {
    let _card = SetCard::try_from(Color::Red | Shade::Empty | Value::Two).unwrap();
  }
  #[test]
  #[should_panic]
  fn test_missing_shade() {
    let _card = SetCard::try_from(Color::Red | Shape::Tilde | Value::Two).unwrap();
  }
  #[test]
  #[should_panic]
  fn test_missing_color() {
    let _card = SetCard::try_from(Shade::Empty | Shape::Tilde | Value::Two).unwrap();
  }
  #[test]
  #[should_panic]
  fn test_missing_value() {
    let _card = SetCard::try_from(Shade::Empty | Shape::Tilde | Color::Green).unwrap();
  }
}
