//! A few macros just to make implemenation easier.
//! In this case, only two macros are defined.
//! One to implement bitwise-and assigning (&=) nad bitside-or assigning (|=) for a generic type ($type) and SetCard.
//! This means that:
//!
//! ```rust
//! use set_card_game::{SetCard, Color};
//! let mut card = SetCard::try_from(0xff).unwrap();
//! // set bits for red color in card
//! card |= Color::Red;
//! ```
//!
//! would be valid code.

macro_rules! impl_bit_operations {
  ($type:ty) => {
    impl std::ops::BitAndAssign<$type> for crate::SetCard {
      fn bitand_assign(&mut self, rhs: $type) {
        self.0 &= rhs as u8
      }
    }
    impl std::ops::BitOr<u8> for $type {
      type Output = u8;
      fn bitor(self, rhs: u8) -> u8 {
        (self as u8) | rhs
      }
    }
    impl std::ops::BitOr<$type> for u8 {
      type Output = u8;
      fn bitor(self, rhs: $type) -> u8 {
        self | (rhs as u8)
      }
    }
    impl std::ops::BitOrAssign<$type> for crate::SetCard {
      fn bitor_assign(&mut self, rhs: $type) {
        self.0 &= rhs as u8
      }
    }
  }
}

/// Implement the equality operator given:
///
/// 1. A value given from some property:
///     * [`crate::Color`]
///     * [`crate::Shape`]
///     * [`crate::Shade`]
///     * [`crate::Value`]
/// 2. A mask value (from [`crate::Property`].
macro_rules! impl_eq {
  ($type:ty, $mask:expr) => {
    impl PartialEq<$type> for crate::SetCard {
      fn eq(&self, other: &$type) -> bool {
        self & $mask == (*other as u8)
      }
    }
  }
}

macro_rules! impl_bit_operations_for_enum {
  ($type:ty) => {
    impl std::ops::BitOr<crate::Color> for $type {
      type Output = u8;

      fn bitor(self, other: crate::Color) -> u8 {
        (self as u8) | (other as u8)
      }
    }
    impl std::ops::BitOr<crate::Shape> for $type {
      type Output = u8;

      fn bitor(self, other: crate::Shape) -> u8 {
        (self as u8) | (other as u8)
      }
    }
    impl std::ops::BitOr<crate::Shade> for $type {
      type Output = u8;

      fn bitor(self, other: crate::Shade) -> u8 {
        (self as u8) | (other as u8)
      }
    }
    impl std::ops::BitOr<crate::Value> for $type {
      type Output = u8;

      fn bitor(self, other: crate::Value) -> u8 {
        (self as u8) | (other as u8)
      }
    }
  }
}
