#[repr(u8)]
#[derive(PartialEq, Eq, Clone, Copy, Debug)]
pub enum Property {
  Color = 0b11,
  Shape = 0b11 << 2,
  Shade = 0b11 << 4,
  Value = 0b11 << 6,
}

/// A set of properties.
///
/// ```rust
/// use set_card_game::{Property, PropertySet};
/// let mut set = PropertySet::empty();
/// // add a new property to the set
/// set |= Property::Color;
/// // check if set contains the color property
/// if set & Property::Color {
///     println!("No color");
/// }
/// # else {
/// # panic!("This condition should never be reached");
/// # }
/// ```
#[derive(Debug)]
pub struct PropertySet(u8);
impl PropertySet {
  pub fn empty() -> Self {
    Self(0)
  }
}

impl PartialEq<u8> for PropertySet {
  fn eq(&self, rhs: &u8) -> bool {
    self.0 == *rhs
  }
}
impl std::ops::BitOr<Property> for PropertySet {
  type Output = PropertySet;

  fn bitor(self, other: Property) -> Self::Output {
    PropertySet(self.0 | (other as u8))
  }
}
impl std::ops::BitOrAssign<Property> for PropertySet {
  fn bitor_assign(&mut self, other: Property) {
    self.0 |= other as u8
  }
}
impl std::ops::BitAnd<Property> for PropertySet {
  type Output = bool;

  fn bitand(self, other: Property) -> Self::Output {
    self.0 & (other as u8) == self.0
  }
}

impl std::ops::BitOr<u8> for Property {
  type Output = u8;

  fn bitor(self, other: u8) -> Self::Output {
    (self as u8) | other
  }
}
impl std::ops::BitOr<Property> for u8 {
  type Output = u8;

  fn bitor(self, other: Property) -> Self::Output {
    self | (other as u8)
  }
}
impl std::ops::BitAnd<u8> for Property {
  type Output = u8;

  fn bitand(self, other: u8) -> Self::Output {
    (self as u8) & other
  }
}
impl std::ops::BitAnd<Property> for u8 {
  type Output = u8;

  fn bitand(self, other: Property) -> Self::Output {
    self & (other as u8)
  }
}

#[repr(u8)]
#[derive(PartialEq, Eq, Clone, Copy, Debug)]
pub enum Color {
  Red = 1,
  Green = 2,
  Purple = 3,
}

#[repr(u8)]
#[derive(PartialEq, Eq, Clone, Copy, Debug)]
pub enum Shape {
  Diamond = 1 << 2,
  Circle = 2 << 2,
  Tilde = 3 << 2,
}

#[repr(u8)]
#[derive(PartialEq, Eq, Clone, Copy, Debug)]
pub enum Shade {
  Solid = 1 << 4,
  Shaded = 2 << 4,
  Empty = 3 << 4,
}

#[repr(u8)]
#[derive(PartialEq, Eq, Clone, Copy, Debug)]
pub enum Value {
  One = 1 << 6,
  Two = 2 << 6,
  Three = 3 << 6,
}

impl_bit_operations_for_enum!(Color);
impl_bit_operations_for_enum!(Shape);
impl_bit_operations_for_enum!(Shade);
impl_bit_operations_for_enum!(Value);

#[cfg(test)]
mod test {
  use super::{Shade, Property, Color, Shape, Value};

  #[test]
  fn test_property_and_color() {
    let value = 0x3;
    assert_eq!(value & Property::Color, Color::Purple as u8);
  }
  #[test]
  fn test_property_and_shape() {
    let value = 0xc;
    assert_eq!(value & Property::Shape, Shape::Tilde as u8);
  }
  #[test]
  fn test_property_and_shade() {
    let value = 0x30;
    assert_eq!(value & Property::Shade, Shade::Empty as u8);
  }
  #[test]
  fn test_property_and_value() {
    let value = 0xc0;
    assert_eq!(value & Property::Value, Value::Three as u8);
  }
  #[test]
  fn test_shade_or() {
    let value = 0x0 | Shade::Empty;
    assert_eq!(value & Property::Shade, Shade::Empty as u8);
  }
  #[test]
  fn test_color_or() {
    let value = 0x0 | Color::Red;
    assert_eq!(value & Property::Color, Color::Red as u8);
  }
  #[test]
  fn test_shape_or() {
    let value = 0x0 | Shape::Tilde;
    assert_eq!(value & Property::Shape, Shape::Tilde as u8);
  }
  #[test]
  fn test_value_or() {
    let value = 0x0 | Value::Two;
    assert_eq!(value & Property::Value, Value::Two as u8);
  }
}
