use crate::{SetCard, error::Error, Property, PropertySet};

pub struct Set((SetCard, SetCard, SetCard));

impl TryFrom<(SetCard, SetCard, SetCard)> for Set {
  type Error = Error;

  fn try_from(cards: (SetCard, SetCard, SetCard)) -> Result<Set, Self::Error> {
    let mut invalid_properties = PropertySet::empty();
    let color = ((cards.0 & Property::Color) + (cards.1 & Property::Color) + (cards.2 & Property::Color)) % 3 == 0;
    if !color {
      invalid_properties |= Property::Color;
    }
    let shape = ((cards.0 & Property::Shape) + (cards.1 & Property::Shape) + (cards.2 & Property::Shape)) % 3 == 0;
    if !shape {
      invalid_properties |= Property::Shape;
    }
    let shade = ((cards.0 & Property::Shade) + (cards.1 & Property::Shade) + (cards.2 & Property::Shade)) % 3 == 0;
    if !shade {
      invalid_properties |= Property::Shade;
    }
    let value = ((cards.0 & Property::Value) % 3 + (cards.1 & Property::Value) % 3 + (cards.2 & Property::Value) % 3) % 3 == 0;
    if !value {
      invalid_properties |= Property::Value;
    }
    if invalid_properties != 0 {
      return Err(Error::InvalidSet(invalid_properties));
    }
    Ok(Set((cards.0, cards.1, cards.2)))
  }
}

#[cfg(test)]
mod tests {
  use crate::{Color, Shape, Shade, Value, SetCard, Set, Property, error::Error};

  #[test]
  fn test_valid_set_all_same() {
    let cards = (
      SetCard::try_from(
        Color::Red |
        Shape::Tilde |
        Shade::Empty |
        Value::One
      ).unwrap(),
      SetCard::try_from(
        Color::Red |
        Shape::Tilde |
        Shade::Empty |
        Value::One
      ).unwrap(),
      SetCard::try_from(
        Color::Red |
        Shape::Tilde |
        Shade::Empty |
        Value::One
      ).unwrap(),
    );
    let _set = Set::try_from(cards).unwrap();
  }
  #[test]
  fn test_valid_set_diff_color() {
    let cards = (
      SetCard::try_from(
        Color::Red |
        Shape::Tilde |
        Shade::Empty |
        Value::One
      ).unwrap(),
      SetCard::try_from(
        Color::Green |
        Shape::Tilde |
        Shade::Empty |
        Value::One
      ).unwrap(),
      SetCard::try_from(
        Color::Purple |
        Shape::Tilde |
        Shade::Empty |
        Value::One
      ).unwrap(),
    );
    let _set = Set::try_from(cards).unwrap();
  }
  #[test]
  fn test_valid_set_diff_shape() {
    let cards = (
      SetCard::try_from(
        Color::Red |
        Shape::Tilde |
        Shade::Empty |
        Value::One
      ).unwrap(),
      SetCard::try_from(
        Color::Red |
        Shape::Circle |
        Shade::Empty |
        Value::One
      ).unwrap(),
      SetCard::try_from(
        Color::Red |
        Shape::Diamond |
        Shade::Empty |
        Value::One
      ).unwrap(),
    );
    let _set = Set::try_from(cards).unwrap();
  }
  #[test]
  fn test_valid_set_diff_shade() {
    let cards = (
      SetCard::try_from(
        Color::Red |
        Shape::Tilde |
        Shade::Solid |
        Value::One
      ).unwrap(),
      SetCard::try_from(
        Color::Red |
        Shape::Tilde |
        Shade::Shaded |
        Value::One
      ).unwrap(),
      SetCard::try_from(
        Color::Red |
        Shape::Tilde |
        Shade::Empty |
        Value::One
      ).unwrap(),
    );
    let _set = Set::try_from(cards).unwrap();
  }
  #[test]
  fn test_valid_set_diff_value() {
    let cards = (
      SetCard::try_from(
        Color::Red |
        Shape::Tilde |
        Shade::Empty |
        Value::One
      ).unwrap(),
      SetCard::try_from(
        Color::Red |
        Shape::Tilde |
        Shade::Empty |
        Value::Two
      ).unwrap(),
      SetCard::try_from(
        Color::Red |
        Shape::Tilde |
        Shade::Empty |
        Value::Three
      ).unwrap(),
    );
    let _set = Set::try_from(cards).unwrap();
  }
  #[test]
  fn test_valid_set_all_diff() {
    let cards = (
      SetCard::try_from(
        Color::Red |
        Shape::Tilde |
        Shade::Empty |
        Value::One
      ).unwrap(),
      SetCard::try_from(
        Color::Green |
        Shape::Diamond |
        Shade::Solid |
        Value::Two
      ).unwrap(),
      SetCard::try_from(
        Color::Purple |
        Shape::Circle |
        Shade::Shaded |
        Value::Three
      ).unwrap(),
    );
    let _set = Set::try_from(cards).unwrap();
  }
  #[test]
  fn test_invaid_set() {
    let cards = (
      SetCard::try_from(
        Color::Red |
        Shape::Tilde |
        Shade::Empty |
        Value::One
      ).unwrap(),
      SetCard::try_from(
        Color::Red |
        Shape::Diamond |
        Shade::Solid |
        Value::Two
      ).unwrap(),
      SetCard::try_from(
        Color::Purple |
        Shape::Circle |
        Shade::Shaded |
        Value::Three
      ).unwrap(),
    );
    let set = Set::try_from(cards);
    match set {
      Err(Error::InvalidSet(e)) => {
        assert!(e & Property::Color);
      }
      Err(_) => panic!("Wrong type of error"),
      Ok(_) => panic!("Produced valid set when dit shouldn't"),
    }
  }
}
